@@ -4,30 +4,43 @@ import json
 from collections import namedtuple
 from distutils.version import StrictVersion
 from functools import wraps
-from typing import Dict, Tuple, Union
+from typing import Dict, Tuple, Union, Optional, List  # noqa
 from unittest.mock import Mock, patch
-from urllib.parse import parse_qsl, urlencode, urlparse
 
-import aiohttp
-from aiohttp import ClientConnectionError, ClientResponse, client, hdrs
+from aiohttp import ClientConnectionError, ClientResponse, ClientSession, hdrs
 from aiohttp.helpers import TimerNoop
 from multidict import CIMultiDict
 
-from .compat import URL, merge_url_params, stream_reader
-
-VERSION = StrictVersion(aiohttp.__version__)
+from .compat import (
+    URL,
+    Pattern,
+    stream_reader_factory,
+    merge_params,
+    normalize_url,
+    AIOHTTP_VERSION
+)
 
 
-class UrlResponse(object):
-    resp = None
+class RequestMatch(object):
+    url_or_pattern = None  # type: Union[URL, Pattern]
 
-    def __init__(self, url: str, method: str = hdrs.METH_GET,
-                 status: int = 200, body: str = '',
+    def __init__(self, url: Union[str, Pattern],
+                 method: str = hdrs.METH_GET,
+                 status: int = 200,
+                 body: str = '',
+                 payload: Dict = None,
                  exception: 'Exception' = None,
-                 headers: Dict = None, payload: Dict = None,
+                 headers: Dict = None,
                  content_type: str = 'application/json',
-                 response_class=None):
-        self.url = self.parse_url(url)
+                 response_class: 'ClientResponse' = None,
+                 timeout: bool = False,
+                 repeat: bool = False):
+        if isinstance(url, Pattern):
+            self.url_or_pattern = url
+            self.match_func = self.match_regexp
+        else:
+            self.url_or_pattern = normalize_url(url)
+            self.match_func = self.match_str
         self.method = method.lower()
         self.status = status
         if payload is not None:
@@ -36,28 +49,31 @@ class UrlResponse(object):
             body = str.encode(body)
         self.body = body
         self.exception = exception
+        if timeout:
+            self.exception = asyncio.TimeoutError('Connection timeout test')
         self.headers = headers
         self.content_type = content_type
         self.response_class = response_class or ClientResponse
+        self.repeat = repeat
 
-    def parse_url(self, url: str) -> str:
-        """Normalize url to make comparisons."""
-        url = str(url)
-        _url = url.split('?')[0]
-        query = urlencode(sorted(parse_qsl(urlparse(url).query)))
+    def match_str(self, url: URL) -> bool:
+        return self.url_or_pattern == url
 
-        return '{}?{}'.format(_url, query) if query else _url
+    def match_regexp(self, url: URL) -> bool:
+        return bool(self.url_or_pattern.match(str(url)))
 
-    def match(self, method: str, url: str) -> bool:
+    def match(self, method: str, url: URL) -> bool:
         if self.method != method.lower():
             return False
-        return self.url == self.parse_url(url)
+        return self.match_func(url)
 
-    def build_response(self) -> Union[ClientResponse, Exception]:
+    async def build_response(
+            self, url: URL
+    ) -> 'Union[ClientResponse, Exception]':
         if isinstance(self.exception, Exception):
             return self.exception
         kwargs = {}
-        if VERSION >= StrictVersion('3.1.0'):
+        if AIOHTTP_VERSION >= StrictVersion('3.1.0'):
             loop = Mock()
             loop.get_debug = Mock()
             loop.get_debug.return_value = True
@@ -65,34 +81,33 @@ class UrlResponse(object):
             kwargs['writer'] = Mock()
             kwargs['continue100'] = None
             kwargs['timer'] = TimerNoop()
-            if VERSION >= StrictVersion('3.3.0'):
+            if AIOHTTP_VERSION >= StrictVersion('3.3.0'):
                 pass
             else:
                 kwargs['auto_decompress'] = True
             kwargs['traces'] = []
             kwargs['loop'] = loop
             kwargs['session'] = None
-        self.resp = self.response_class(self.method, URL(self.url), **kwargs)
+        resp = self.response_class(self.method, url, **kwargs)
         # we need to initialize headers manually
         headers = CIMultiDict({hdrs.CONTENT_TYPE: self.content_type})
         if self.headers:
             headers.update(self.headers)
         raw_headers = self._build_raw_headers(headers)
-        if VERSION >= StrictVersion('3.3.0'):
+        if AIOHTTP_VERSION >= StrictVersion('3.3.0'):
             # Reified attributes
-            self.resp._headers = headers
-            self.resp._raw_headers = raw_headers
+            resp._headers = headers
+            resp._raw_headers = raw_headers
         else:
-            self.resp.headers = headers
-            self.resp.raw_headers = raw_headers
-        self.resp.status = self.status
-        self.resp.content = stream_reader()
-        self.resp.content.feed_data(self.body)
-        self.resp.content.feed_eof()
+            resp.headers = headers
+            resp.raw_headers = raw_headers
+        resp.status = self.status
+        resp.content = stream_reader_factory()
+        resp.content.feed_data(self.body)
+        resp.content.feed_eof()
+        return resp
 
-        return self.resp
-
-    def _build_raw_headers(self, headers):
+    def _build_raw_headers(self, headers: Dict) -> Tuple:
         """
         Convert a dict of headers to a tuple of tuples
 
@@ -104,10 +119,14 @@ class UrlResponse(object):
         return tuple(raw_headers)
 
 
+RequestCall = namedtuple('RequestCall', ['args', 'kwargs'])
+
+
 class aioresponses(object):
     """Mock aiohttp requests made by ClientSession."""
-    _responses = None
-    method_call = namedtuple('method_call', ['args', 'kwargs'])
+    _matches = None  # type: List[RequestMatch]
+    _responses = None  # type: List[ClientResponse]
+    requests = None  # type: Dict
 
     def __init__(self, **kwargs):
         self._param = kwargs.pop('param', None)
@@ -146,47 +165,54 @@ class aioresponses(object):
                     return f(*args, **kwargs)
         return wrapped
 
+    def clear(self):
+        self._responses.clear()
+        self._matches.clear()
+
     def start(self):
         self._responses = []
+        self._matches = []
         self.patcher.start()
         self.patcher.return_value = self._request_mock
 
     def stop(self) -> None:
-        for r in self._responses:
-            if r.resp is not None:
-                r.resp.close()
+        for response in self._responses:
+            response.close()
         self.patcher.stop()
-        self._responses = []
+        self.clear()
 
-    def head(self, url: str, **kwargs):
+    def head(self, url: 'Union[URL, str]', **kwargs):
         self.add(url, method=hdrs.METH_HEAD, **kwargs)
 
-    def get(self, url: str, **kwargs):
+    def get(self, url: 'Union[URL, str]', **kwargs):
         self.add(url, method=hdrs.METH_GET, **kwargs)
 
-    def post(self, url: str, **kwargs):
+    def post(self, url: 'Union[URL, str]', **kwargs):
         self.add(url, method=hdrs.METH_POST, **kwargs)
 
-    def put(self, url: str, **kwargs):
+    def put(self, url: 'Union[URL, str]', **kwargs):
         self.add(url, method=hdrs.METH_PUT, **kwargs)
 
-    def patch(self, url: str, **kwargs):
+    def patch(self, url: 'Union[URL, str]', **kwargs):
         self.add(url, method=hdrs.METH_PATCH, **kwargs)
 
-    def delete(self, url: str, **kwargs):
+    def delete(self, url: 'Union[URL, str]', **kwargs):
         self.add(url, method=hdrs.METH_DELETE, **kwargs)
 
-    def options(self, url: str, **kwargs):
+    def options(self, url: 'Union[URL, str]', **kwargs):
         self.add(url, method=hdrs.METH_OPTIONS, **kwargs)
 
-    def add(self, url: str, method: str = hdrs.METH_GET, status: int = 200,
+    def add(self, url: 'Union[URL, str]', method: str = hdrs.METH_GET,
+            status: int = 200,
             body: str = '',
             exception: 'Exception' = None,
             content_type: str = 'application/json',
             payload: Dict = None,
             headers: Dict = None,
-            response_class=None) -> None:
-        self._responses.append(UrlResponse(
+            response_class: 'ClientResponse' = None,
+            repeat: bool = False,
+            timeout: bool = False) -> None:
+        self._matches.append(RequestMatch(
             url,
             method=method,
             status=status,
@@ -196,43 +222,44 @@ class aioresponses(object):
             payload=payload,
             headers=headers,
             response_class=response_class,
+            repeat=repeat,
+            timeout=timeout,
         ))
 
-    def match(self, method: str, url: str) -> 'ClientResponse':
-        i, resp = next(
-            iter(
-                [(i, r.build_response())
-                 for i, r in enumerate(self._responses)
-                 if r.match(method, url)]
-            ),
-            (None, None)
-        )
+    async def match(self, method: str, url: URL) -> Optional['ClientResponse']:
+        for i, matcher in enumerate(self._matches):
+            if matcher.match(method, url):
+                response = await matcher.build_response(url)
+                break
+        else:
+            return None
 
-        if i is not None:
-            del self._responses[i]
-        if isinstance(resp, Exception):
-            raise resp
-        return resp
+        if matcher.repeat is False:
+            del self._matches[i]
+        if isinstance(response, Exception):
+            raise response
+        return response
 
-    async def _request_mock(self, orig_self: client.ClientSession,
-                            method: str, url: str, *args: Tuple,
+    async def _request_mock(self, orig_self: ClientSession,
+                            method: str, url: 'Union[URL, str]',
+                            *args: Tuple,
                             **kwargs: Dict) -> 'ClientResponse':
         """Return mocked response object or raise connection error."""
-
-        url = merge_url_params(url, kwargs.get('params'))
-
+        url = normalize_url(merge_params(url, kwargs.get('params')))
+        url_str = str(url)
         for prefix in self._passthrough:
-            if str(url).startswith(prefix):
+            if url_str.startswith(prefix):
                 return (await self.patcher.temp_original(
                     orig_self, method, url, *args, **kwargs
                 ))
 
-        response = self.match(method, url)
+        response = await self.match(method, url)
         if response is None:
             raise ClientConnectionError(
                 'Connection refused: {} {}'.format(method, url)
             )
+        self._responses.append(response)
         key = (method, url)
-        self.requests.setdefault(key, list())
-        self.requests[key].append(self.method_call(args, kwargs))
+        self.requests.setdefault(key, [])
+        self.requests[key].append(RequestCall(args, kwargs))
         return response
